# Cardigan

Cardigan offers a library to both run the Cardigan algorithm and handle multiple PPI networks, and disease-gene interaction sources (such as OMIM)

## Getting Started

These instructions will show how to setup Cardigan and allow its usage as a Python module.

### Prerequisites

Cardigan requires the installation of NumPY and SciPY:

```
$ pip install scipy
```

Furthermore, you need to provide at least one PPI network, a Caniza disease similarity matrix, and both OMIM morbidmap and OMIM mim2gene datasets. The datasets are located by default under the `data` folder in default folder scheme:

```bash
   .
    ├── bin                 # Executables and examples
    ├── data                # Data files
    │   ├── dissim          # Caniza similarity
    │   ├── mappings
    │   │   ├── diamond     # disease     
    │   │   └── omim        # morbidmap and mim2gene
    │   └── ppi              
    │       ├── biogrid     # BioGRID (human)
    │       ├── diamond     # DiamondNet
    │       ├── funcoup     # FUNCOUP 3.0 (human)
    │       ├── hippie      # HIPPIE
    │       ├── hprd        # HPRD
    │       └── user        # Custom network
    ├── lib                 # Source files
    ├── LICENSE
    └── README.md
```


The Caniza disease similarity calculated with the 2017 OMIM is provided in the [data bundle](paccanarolab.org/cardigan), otherwise it must be computed by the user. The software to produce the matrix can be downloaded [here](paccanarolab.org/disease_similarity).

At least one of the networks (found under `data/ppi`) should be provided. BioGRID in TAB2 format, and DiamondNet from the Ghiassian *et. al.* paper (S1 file), FUNCOUP v3.0, HIPPIE and HPRD in their standard format. The custom PPI network should be a tab separated file with Entrez identifiers (called `ppi.txt` as default), and may include a third column with the edge weight (all edges should include weights if the PPI is weighted). The default network can be established in the `setPPI(ppiType=)` method found `lib/Config.py` (it defaults to `'custom'`).

By default, we expect to find the following files in the `data` folder:

```bash
   .
    ├── dissim
    │   └── 2017-CanizaSimilarity.out
    ├── mappings 
    │   ├── diamond # optional
    │   │   └── 201504-18_PLoS-DIAMOnD-S2-Data.txt
    │   └── omim
    │       ├── morbidmap.txt
    │       └── mim2gene.txt
    └── ppi # at least 1 of the below folders
        ├── biogrid     
        │   └── BIOGRID-ORGANISM-Homo_sapiens-x.y.z.tab2.txt
        ├── diamond     
        │   └── 201504-18_PLoS-DIAMOnD-S1-Data.txt
        ├── funcoup     
        │   └── FC3.0_H.sapiens_full
        ├── hippie      
        │   └── hippie.txt
        ├── hprd        
        │   └── BINARY_PROTEIN_PROTEIN_INTERACTIONS.txt
        └── user        
            └── ppi.txt

```

All file names and locations can be changed in `lib/Config.py`.

### Installing

The modules can run as they are, however, the Caniza similarity must be pre-processed for faster access within the method. The pre-process and data availability can be done launching the setup:

```
$ python setup.py
```

If the setup is not executed, the Caniza similarity will be pre-processed the first time Cardigan runs.

## Running Cardigan

You can find working examples under the `bin` folder. 

The following examples breakdown how Cardigan can be used to run disease predictions using all the data available in OMIM, and a synthetic prediction (where some gene associations are removed):

### Minimum working example

This is the standard usage of our software

```python
import Cardigan

# load the module using the default configuration
cardigan = Cardigan.Cardigan()

# predict genes for BDPLT16 (MIM: 187800)
out = cardigan.predict(['187800'])
print out

```

### Synthetic leave-one-out

In this example we remove a single known gene from a disease, and calculate the prediction

```python
import Cardigan

# load the module using the default configuration
cardigan = Cardigan.Cardigan()

# predict genes for BDPLT16 (MIM: 187800)
# keep gene 3690 as a seed and use 3674 as a target
out = cardigan.predict(['187800'],['3690'],['3674'])
targetPos = cardigan.evaluate()
print targetPos

```

### Changing the PPI network

You can establish the network by loading the appropriate `config` to Cardigan:

```python
import Cardigan
import Config

# load the configuration module
config = Config.Config()
config.setPPI('biogrid')
# load the module with the modified configuration
cardigan = Cardigan.Cardigan(config)

# predict genes for BDPLT16 (MIM: 187800)
# keep gene 3690 as a seed and use 3674 as a target
out = cardigan.predict(['187800'],['3690'],['3674'])
targetPos = cardigan.evaluate()
print targetPos

```


## Author

* **Juan Caceres**


## License

This project is licensed under the CC BY-NC 4.0 International License - see the [LICENSE.md](LICENSE.md) file for details

