"""
    Cardigan - Disease Gene Prediction Method
    Copyright (C) 2018  Juan Caceres

    Cardigan is licensed under a Creative Commons Attribution-NonCommercial 4.0 
    International License.  If you did not get a copy of the license with
    the source code, you can get a copy at:
    
    https://creativecommons.org/licenses/by-nc/4.0/legalcode

"""
__author__  = "Juan Caceres"
__email__   = "Juan.CaceresSilva.2014@live.rhul.ac.uk"
__copyright__ = "Copyright (C) 2018 Juan Caceres"
__license__ = "CC BY-NC"
__version__ = "1.0"

from collections import defaultdict
from itertools import chain
import random

import Config
import PPIUtilities
import OMIMUtilities
import DISSIMUtilities
import diffusionUtilities
import Measures

import abc
import os
import math
import numpy as np

class Method(object):
	__metaclass__ = abc.ABCMeta

	lastPrediction = None

	@abc.abstractmethod
	def predict(self, seedMims, seedGenes, targetGenes, minTargets):
		'''
		Considering the diffusion can be modelled as Ax = b, where A is the 
		diffusion matrix, and b is the initial guess vector.

		This method must return the resulting vector x (as numpy array).

		*The diffuse method should run without parameters, they should be
		set in the setParameters() method.
		'''
	
	@abc.abstractmethod
	def getName(self):
		'''
		Returns the name of the diffusion method
		'''

	def __str__(self):
		return self.getName()

class Cardigan(Method):


	def __init__(self, config=False, outSize=250, verbose=True):
		if config:
			self.config = config
		else:
			#default parameters of current config
			self.config = Config.Config()

		#Parameters for the sigmoid tuning 
		self.sigmoidHeight = self.config.sigmoidHeight #0.5
		self.sigmoidTen = self.config.sigmoidTen #3.0
		self.sigmoidNinety = self.config.sigmoidNinety #4.0

		try:
			ppi = self.config.ppiObject
		except:
			self.config.setPPI()
			ppi = self.config.ppiObject

		try:
			#load the disease module handler
			self.modules = OMIMUtilities.DiseaseModules(self.config.diseaseModules,self.config.diseaseModulesMims)
		except:
			pass

		mim2gene = OMIMUtilities.Mim2Gene(self.config.mim2gene)
		morbidMap =  OMIMUtilities.MorbidMap(self.config.morbidMap,separator=self.config.morbidMapSep,mim2gene=mim2gene)

		dissim = self.config.dissim
		dissimParsed = self.config.dissimParsed

		diseaseSimilarities = DISSIMUtilities.DiseaseSimilarity(dissim, outfile=dissimParsed)
	

		self.diffusion = diffusionUtilities.Diffusion(ppi, morbidMap, diseaseSimilarities,verbose)
		self.__diffusionType__ = 'consistency'
		self.diffusion.setSigmoidScale(self.sigmoidTen, self.sigmoidNinety, self.sigmoidHeight)
		self.diffusion.setDiffusionParameters(self.__diffusionType__,alpha=self.config.diffusionAlpha)

		self.__outSize__ = outSize
		self.__verbose__ = verbose

	def predict(self, seedMims, seedGenes='default', targetGenes=[], minTargets=0):
		'''
		Executes a prediction with our method

		seedMims: is a list of OMIM identifiers (elements are strings)

		seedGenes: can be 'default' to obtain the seeds from the disease-gene mapping database,
		or a list of Entrez identifiers as strings.

		targetGenes: is a list of Entrez identifiers to be predicted by the method.

		minTargets is an int establishing the minimum amount of targets needed
		before returning a void prediction.
		'''
		if len(set(targetGenes)&set(self.diffusion.ppi.vertices))<minTargets:
			return None


		self.diffusion.setDiffusionMims(seedMims)
		if seedGenes!='default':
			self.diffusion.setDiffusionGenes(seedGenes)
		if len(targetGenes)>0:
			self.diffusion.setDiffusionTargets(targetGenes)

		result = self.diffusion.diffuse()

		if result:
			self.lastPrediction =  result.getFirstNGenes(self.__outSize__, ignoreSeeds = True)
			#self.lastPrediction =  self.diffusion.predictionResult.getFirstNGenes(self.__outSize__, ignoreSeeds = True)

			return self.lastPrediction
		else:
			return None

	def evaluate(self,measure=None,measureParam=None):
		'''
		Evaluate the last prediction

		-measure: is an optional parameter to override the config measure. 
		The possible values are:
		  'roc': (higher is better) value of the ROC AUC defined by the positions of the targets 
		  		in the prediction vector 
		  'mean': (lower is better) average position of the targets in the prediction vector
		  'recall': (higher is better) returns 1 if the target is within the top X values (X is 
		  		a parameter of the evaluation)
		'''
		if not measure:
			measure=self.config.measure
		if not measureParam:
			measureParam=self.config.measureParam
		
		return self.diffusion.evaluatePredictions(measure=measure,measureParam=measureParam)

	def getName(self):
		'''Returns the name of the method

		includeParams is a boolean to include relevant params within the name
		'''
		return 'Cardigan'+'('+self.diffusion.ppi.name+')'
