"""
    Cardigan - Disease Gene Prediction Method
    Copyright (C) 2018  Juan Caceres

    Cardigan is licensed under a Creative Commons Attribution-NonCommercial 4.0 
    International License.  If you did not get a copy of the license with
    the source code, you can get a copy at:
    
    https://creativecommons.org/licenses/by-nc/4.0/legalcode

"""
__author__  = "Juan Caceres"
__email__   = "Juan.CaceresSilva.2014@live.rhul.ac.uk"
__copyright__ = "Copyright (C) 2018 Juan Caceres"
__license__ = "CC BY-NC"
__version__ = "1.0"

import os
import PPIUtilities

class Config:

	def __init__(self, **kwargs):
		'''
		Optional arguments are:

		-dataDir: base directory for all data files (OMIM databases, PPIs, disease similarity)

		-morbidMap: location of the OMIM morbidmap file.

		-morbidMapSep: column separator of the OMIM morbidmap (as this file is produced,
		the default separator is a TAB, and will fallback to the legacy | ).

		-mim2gene: location of the OMIM mim2gene file.

		-diamondModules: location of the DIAMOnD S2 file.

		-diamondModulesMims: mapping between the DIAMOnD diseases to OMIM diseases.

		-dissim: raw output of the Caniza similarity.

		-dissimParsed: shelf version of the Caniza similarity produced by our DISSIMUtilities 
		module.

		'''

		self.sigmoidHeight = 0.5
		self.sigmoidTen=3.0
		self.sigmoidNinety=4.0
		self.diffusionAlpha=0.69
		#the measure can be: 
		# - 'roc': (higher is better) value of the ROC AUC defined by the positions of the targets in the prediction vector 
		# - 'mean': (lower is better) average position of the targets in the prediction vector
		# - 'recall': (higher is better) returns 1 if the target is within the top X values (X is a parameter of the evaluation)
		self.measure='mean'

		self.measureParam=200
		self.disease2gene='morbidmap'

		try:
			self.__dataDir__=kwargs['dataDir']
		except:
			self.__dataDir__=os.path.normpath(os.path.join('..','data'))

		##################
		### MorbidMaps ###
		##################
		self.mmDefault={'file':os.path.join(self.__dataDir__,'mappings','omim','morbidmap.txt'),'sep':'\t'}
		#self.mmDefalut={'file':os.path.join(self.__dataDir__,'omim','morbidmap.txt'),'sep':'|'}
		self.mmOldFormat={'file':os.path.join(self.__dataDir__,'mappings','omim','morbidmap.txt'),'sep':'|'}
		try:
			self.mim2gene=os.path.join(self.__dataDir__,kwargs['mim2gene'])
		except:
			self.mim2gene=os.path.join(self.__dataDir__,'mappings','omim','mim2gene.txt')

		try:
			self.morbidMap=os.path.join(self.__dataDir__,kwargs['morbidMap'])
			self.morbidMapSep=kwargs['morbidMapSep']
		except:
			self.morbidMap=self.mmDefault['file']
			self.morbidMapSep=self.mmDefault['sep']
			

		########################
		### Disease Families ###
		########################
		self.diamondDiseaseModules = os.path.join(self.__dataDir__,'mappings','diamond','201504-18_PLoS-DIAMOnD-S2-Data.txt')
		self.diamondDiseaseModulesMims = os.path.join(self.__dataDir__,'mappings','diamond','diseasesToMIM.txt')
		try:
			self.diseaseModules = os.path.join(self.__dataDir__,kwargs['diseaseModules'])
			self.diseaseModulesMims = os.path.join(self.__dataDir__,kwargs['diseaseModulesMims'])
		except:
			self.diseaseModules = self.diamondDiseaseModules
			self.diseaseModulesMims = self.diamondDiseaseModulesMims


		############################
		### Disease Similarities ###
		############################
		#The raw similarity data is only used to produce the disease similarity shelf.

		#Raw output of the Caniza method ()
		self.dissim2017=self.dissim=os.path.join(self.__dataDir__,'dissim','2017-CanizaSimilarity.out')
		#Python shelf version to avoid full load of the data (produced by our DISSIMUtilities module)
		self.dissimParsed2017=os.path.join(self.__dataDir__,'dissim','2017-CanizaSimilarity.dissim')
		

		try:
			self.dissim=os.path.join(self.__dataDir__,kwargs['dissim'])
		except:
			self.dissim=self.dissim2017

		try:
			self.dissimParsed=os.path.join(self.__dataDir__,kwargs['dissimParsed'])
		except:
			self.dissimParsed=self.dissimParsed2017


		####################
		### PPI networks ###
		####################

		# Our PPI network module provides a framework to load networks which
		# require identifier translations or may contain primary and secondary
		# identifiers (used as fallback when the primary identifier is
		# missing). Refer to our PPIUtilities module for further details.
		try:
			self.ppiObject = kwargs['ppiObject']
		except:
			pass


	def setPPI(self,ppiType='custom',createObject=True):
		'''ppiType can be one of the following predefined values:
		- hprd: Standard HPRD file.
		- biogrid: BioGRID interactome in the tab2 format.
		- hippie: Standard HIPPIE file.
		- funcoup: Standard FUNCOUP file.
		- diamondnet: Interactome in the DIAMOnD paper (S1 file).
		- custom: Custom PPI. Each row is a pair of Entrez identifiers separated by a TAB.
		'''
		#if you want to include a new PPI, check the PPIUtilities file
		self.ppiType=ppiType.lower()
		self.ppiName=ppiType.upper()
		self.ppiColumns=ppiType.lower()
		if ppiType.lower()=='hprd':
			thisDir = os.path.join(self.__dataDir__,'ppi','hprd')
			self.ppi=os.path.join(thisDir,'BINARY_PROTEIN_PROTEIN_INTERACTIONS.txt')
			#symbol to entrez identifier map obtained from HGNC, you can replace this file if needed
			#the file has a symbol and an Entrez identifier per line separated by a TAB
			self.translateFile=os.path.join(thisDir,'symbol_to_entrez.txt')
			#what are the columns in the translation file
			self.translateCols=['source','target']
			#identifier found in the PPI to be replaced
			#our method works using Entrez
			self.translateIdentifier='symbol'

		elif ppiType.lower()=='diamondnet':
			thisDir = os.path.join(self.__dataDir__,'ppi','diamond')
			self.ppi=os.path.join(thisDir,'201504-18_PLoS-DIAMOnD-S1-Data.txt')
			self.ppiName='diamondNet'

		elif ppiType.lower()=='biogrid':
			thisDir = os.path.join(self.__dataDir__,'ppi','biogrid')
			self.ppi=os.path.join(thisDir,'BIOGRID-ORGANISM-Homo_sapiens-x.y.z.tab2.txt')
			self.ppiName='BioGRID'

		elif ppiType.lower()=='hippie':
			thisDir = os.path.join(self.__dataDir__,'ppi','hippie')
			self.ppi=os.path.join(thisDir,'hippie.txt')

		elif ppiType.lower()=='funcoup':
			thisDir = os.path.join(self.__dataDir__,'ppi','funcoup')
			self.ppi=os.path.join(thisDir,'FC3.0_H.sapiens_full')
			#entrez to ensembl identifier map obtained from HGNC, you can replace this file if needed
			#the file has an Entrez identifier and an Ensembl identifier per line separated by a TAB
			self.translateFile=os.path.join(thisDir,'funcoup','entrez_to_ensembl.txt')
			#what are the columns in the translation file
			self.translateCols=['target','source']
			#identifier found in the PPI to be replaced
			#our method works using Entrez
			self.translateIdentifier='ensembl'
		else:

			self.ppi=os.path.join(self.__dataDir__,'ppi','user','ppi.txt')
			#the custom PPI will obtain its name from the file
			ppiName=os.path.basename(self.ppi)			
			self.ppiName=ppiName.split('.')[0]
			self.ppiColumns='entrez_pair'

		if ppiType.lower()=='funcoup' or ppiType.lower()=='hippie':		
			self.sigmoidHeight = 0.6
			self.sigmoidTen=3.0
			self.sigmoidNinety=4.0
			self.diffusionAlpha=0.75
		else:
			self.sigmoidHeight = 0.5
			self.sigmoidTen=3.0
			self.sigmoidNinety=4.0
			self.diffusionAlpha=0.69

		# If the default name has not been set, check for any extra files in
		# the PPI folder. If many many files are provided, the first listed
		# file will be picked
		if not os.path.isfile(self.ppi):
			try:
				# these files are provided by our bundle, and are not PPI networks
				auxiliaryFiles = ['entrez_to_ensembl.txt','symbol_to_entrez.txt']
				file = [f for f in os.listdir(thisDir) if os.path.isfile(os.path.join(thisDir, f)) and f not in auxiliaryFiles][0]
				self.ppi = os.path.join(thisDir,file)
			except:
				print 'Could not find the PPI file in',thisDir

		# see if the custom PPI is binary or weighted
		if ppiType.lower() == 'custom':
			with open(self.ppi,'r') as f:
				for line in f:
					if len(line.split('\t'))==3:
						self.ppiColumns='entrez_pair_weighted'
					break


		if createObject:
			if ppiType.lower()=='funcoup' or ppiType.lower()=='hprd':
				self.ppiObject = PPIUtilities.PPI(self.ppi, columns=self.ppiColumns, ppiName=self.ppiName,
					mainIdentifier=self.translateIdentifier, translateFile=self.translateFile,
					translateCols=self.translateCols)
			else:
				self.ppiObject = PPIUtilities.PPI(self.ppi, columns=self.ppiColumns, ppiName=self.ppiName)
