"""
    Cardigan - Disease Gene Prediction Method
    Copyright (C) 2018  Juan Caceres

    Cardigan is licensed under a Creative Commons Attribution-NonCommercial 4.0 
    International License.  If you did not get a copy of the license with
    the source code, you can get a copy at:
    
    https://creativecommons.org/licenses/by-nc/4.0/legalcode

"""
__author__  = "Juan Caceres"
__email__   = "Juan.CaceresSilva.2014@live.rhul.ac.uk"
__copyright__ = "Copyright (C) 2018 Juan Caceres"
__license__ = "CC BY-NC"
__version__ = "1.0"

from collections import defaultdict
from tsvParser import Parser
import shelve
import os.path
import pickle
import numpy as np
from scipy.spatial.distance import pdist, squareform
from datetime import datetime
import gzip

class DiseaseSimilarity:
	'''
	Object to handle disease similarities.
	'''
	def __init__(self, filename, outfile = None, columns = None, rebuild=False, verbose = True,
		source = '3col',preload=[]):
		'''
		filename is the path to a tsv file that must include the columns
			mim_a, mim_b, similarity

		outfile is the path of the parsed similarity matrix to be used. This
		file will be created if non existent, or used if found (avoiding re-
		creation and lowering memory usage).

		columns is the list of columns to be used by the parser. Columns must
		include ['mim_a','mim_b','similarity'] if used

		rebuild is a boolean flag to rebuild the output file even if found

		verbose is a boolean flag to verbose output

		source is the type of input file. The default option is a tsv that
		includes ['mim_a','mim_b','similarity']. The other option is 'matrix',
		that loads a  full tsv matrix where the first column is the disease
		identifier (van Driel - MimMimer format).

		preload is a list of mims to preload if required. We will not use the
		shelf if this argument is given.

		'''

		self.__filename__ = filename
		if not columns:
			self.__columns__ = ['mim_a','mim_b','similarity']
		else:
			self.__columns__ = columns

		if not outfile:
			self.__outfile__ = datetime.now().strftime("%s")+'.dissim.tmp'
		else:
			self.__outfile__ = outfile

		self.__verbose__ = verbose

		#if preload:
		#	self.loadAndPreload(preload)
		#else:
		#check if files exist_
		fileExists = (os.path.isfile(self.__outfile__) and os.path.isfile(self.__outfile__+'.pkl'))
		
		#create files if they don't exist or a rebuild is forced
		if rebuild or not fileExists:
			if self.__verbose__:
				print 'Creating shelve'
			if source == 'matrix':
				self.__similarities__, self.mims = self.parseMatrix()
			else:
				#start tsvParser Parser object on the given file
				self.__parser = Parser(self.__filename__, self.__columns__,strict=True)
				#load data from the file
				self.__similarities__, self.mims = self.parse()

		else:
			#open the shelve
			self.__similarities__ = shelve.open(self.__outfile__)
			#open the pickle
			mimPkl = open(self.__outfile__+'.pkl')
			#load data
			self.mims = pickle.load(mimPkl)
			#close the pickle
			mimPkl.close()

	def parse(self):
		'''
		Creates a disease-disease -> similarity dictionary from a 3 column TSV
		MIM_a\tMIM_b\tsimilarity

		disease is a MIM
		similarity is a real number (0 to 4 approx)

		filename is the path to the combined similarity matrix
		'''
		dictionary = shelve.open(self.__outfile__)
		mims = set()
		for parsedLine in self.__parser:
		
			try:
				key = (min(parsedLine.mim_a,parsedLine.mim_b),max(parsedLine.mim_a,parsedLine.mim_b))
				key = str(key)
				dictionary[key] = float(parsedLine.similarity)
				mims|={parsedLine.mim_a,parsedLine.mim_b}
			except:
				continue

		if self.__verbose__:
			print 'Found',len(mims),'diseases and',len(dictionary),'non zero similarities'

		#pickle and save mim dictionary
		mimPkl = open(self.__outfile__+'.pkl', 'wb')
		#dump
		pickle.dump(mims,mimPkl)
		#close the pickle
		mimPkl.close()

		return dictionary,mims

	def getDiseases(self):
		'''
		Retrieves all the disease MIMs that have similarities
		'''
		return self.mims


	def similarity(self, mimA, mimB):
		'''
		Retrieves the similarity between diseases A and B
		'''

		#order the pair
		mimPair = str((min(mimA,mimB),max(mimA,mimB)))

		if mimPair in self.__similarities__:
			return self.__similarities__[mimPair]
		else:
			#pairs with a similarity value of zero are not stored, to save space
			return 0

	def parseMatrix(self):
		'''
		Creates a disease-disease -> similarity dictionary from a similarity matrix
		
		MIM_1\tsimilarity_1\t...\tsimilarity_n
		.
		.
		.
		MIM_n\tsimilarity_1\t...\tsimilarity_n

		disease is a MIM
		similarity is a real number

		filename is the path to the combined similarity matrix
		'''

		dictionary = shelve.open(self.__outfile__)
		mims = list()

		#get the disease mims available in the matrix
		with gzip.open(self.__filename__,'r') as f:

			for line in f:
				line = line.strip().split()

				mim = line[0].strip()
				mims.append(mim)

		#start building the pairs
		with gzip.open(self.__filename__,'r') as f:
			for line in f:
				line = line.strip().split()

				mim_a = line[0]
				similarities = line[1:]

				#for this mim, add the pairing with all other diseases
				for e,similarity in enumerate(similarities):

					#ignore null similarity values
					if similarity == '0':
						continue

					mim_b = mims[e]

					key = (min(mim_a,mim_b),max(mim_a,mim_b))
					key = str(key)

					dictionary[key] = float(similarity)

		if self.__verbose__:
			print 'Found',len(mims),'diseases and',len(dictionary),'non zero similarities'

		mims = set(mims)

		#pickle and save mim dictionary
		mimPkl = open(self.__outfile__+'.pkl', 'wb')
		#dump
		pickle.dump(mims,mimPkl)
		#close the pickle
		mimPkl.close()

		

		return dictionary,mims

	def loadAndPreload(self,mims):
		'''
		Similarity matrix from file for the given mims
		'''
		availableMims = set()
		mimDict=dict()
		print 'load indices'
		#get the disease mims available in the matrix
		with open(self.__filename__,'r') as f:

			for line in f:
				line = line.strip().split()

				if line[0] in mims and line[1] in mims:
					mimDict[(line[0],line[1])]=float(line[2])
					availableMims.add(line[0])


		#mims=sorted(set(mims)&availableMims)

		self.__preloadedArray__ = np.zeros((len(mims),len(mims)))
		self.__preloadedMims__ = sorted(availableMims)

		print 'now load matrix',len(mims),self.__preloadedArray__.shape
		'''
		mimDict=dict()
		#get the data
		with open(self.__filename__,'r') as f:
			for line in f:
				line = line.strip().split()

				if line[0] in mims and line[1] in mims:
					mimDict[(line[0],line[1])]=line[2]
					ixA = mims.index(line[0])
					ixB = mims.index(line[1])
					self.__preloadedArray__[ixA][ixB] = line[2]
					self.__preloadedArray__[ixB][ixB] = line[2]

				except:
					pass

		print 'loaded'
		'''				

	def preloadSimilarities(self,mims):
		'''
		Load a numpy array with similarities and closes the pickle
		'''
		mims = sorted(set(mims)&self.mims)

		sims = np.zeros((len(mims),len(mims)));

		for i,mimA in enumerate(mims):
			for j,mimB in enumerate(mims[i:]):
				if mimA in self.mims and mimB in self.mims:
					try:
						sims[i][j]=self.__similarities__[str((mimA,mimB))]
						sims[j][i]=self.__similarities__[str((mimA,mimB))]
					except:
						pass

		self.__preloadedArray__ = sims
		self.__preloadedMims__ = mims

		self.__similarities__.close()

		print 'similarities preloaded'

	def getSimVector(self,mims=[],ignoreMims=[]):
		'''
		Get the similarity vectors for a list of mims

		mims is a list of mim ids

		ignoreMims is a list of mim ids (converted to zero in the output)
		'''

		out = np.zeros((len(mims),len(self.__preloadedMims__)))
		for e,mim in enumerate(mims):
			if mim in self.__preloadedMims__:
				out[e]=self.__preloadedArray__[self.__preloadedMims__.index(mim)]

		for mim in ignoreMims:
			if mim in self.__preloadedMims__:
				out[:,self.__preloadedMims__.index(mim)]=0

		return out
