"""
    Cardigan - Disease Gene Prediction Method
    Copyright (C) 2018  Juan Caceres

    Cardigan is licensed under a Creative Commons Attribution-NonCommercial 4.0 
    International License.  If you did not get a copy of the license with
    the source code, you can get a copy at:
    
    https://creativecommons.org/licenses/by-nc/4.0/legalcode

"""
__author__  = "Juan Caceres"
__email__   = "Juan.CaceresSilva.2014@live.rhul.ac.uk"
__copyright__ = "Copyright (C) 2018 Juan Caceres"
__license__ = "CC BY-NC"
__version__ = "1.0"

import numpy as np

'''
Functions to measure the result of ranking methods
'''


class VectorROC:
	def __init__(self, important=100, verbose=False):
		self.__verbose__=verbose

		self.important=important

	def evaluate2(self, vec = [], buckets = 0):
		'''
		Calculates the ROC of a vector

		vec is a list, it can be of two types:
			- Simple list of booleans, sorted according to the likelihood.
				i.e: element 0 is the most likely, element 1 is the next, ...

			- List of pairs, when the identifiers come with a matching likelihood value.
				i.e.: [[True,0.5],[False,0.9], ... ]

		buckets is the resolution of the false positive axis. It defines the number
		of divisions. A value of 0 means to have as many buckets as input values.
		'''

		if len(vec)<1:
			return

		# check if sorting is needed	
		try:
			if len(vec[0]) == 2:
				vec.sort(key=lambda x: x[1])
			vec = [x[0] for x in vec]
		except:
			pass

		# number of true labels
		trueCount = len([x for x in vec if x])
		# percent of true positive increase
		yDelta = 1.0/trueCount

		#define the amount of false positives per bucket
		if buckets > 0:
			if buckets > len(vec)-trueCount:
				#max amount of buckets
				buckets = len(vec)-trueCount

			bucketSize = (len(vec) - trueCount) / float(buckets)
			if self.__debug__:
				print 'bucket size:',bucketSize
		else:
			buckets = len(vec)-trueCount
			bucketSize = 1


		# percent of false positive increase
		xDelta = 1.0/(buckets)

		#resulting ROC curve
		rocVector = [0]
		#index for bucket jumping
		falseCount = 0


		for x in vec:
			try:
				x = x[0]
			except:
				pass

			if x:
				# true positive found, go up
				rocVector[-1]+=yDelta
			else:
				# false positive found
				falseCount += 1
				if not (
						((int)(falseCount / bucketSize) == (int)((falseCount-1) / bucketSize)) 
						or (len(rocVector) == buckets)
					):
					print falseCount
					# if a new bucket should be made, go left
					rocVector.append(rocVector[-1])


		roc = sum([x*xDelta for x in rocVector])

		if self.__verbose__:
			#print rocVector

			print 'ROC:',roc,'for',len(rocVector),'buckets, from',len(vec),'elements and',trueCount,'true labels'

		return roc

	def evaluate(self, positions, totalElements, buckets=0, important=False):
		'''
		Calculates the ROC from a vector of true label positions

		positions is a list with the positions of the predicted elements.

		totalElements is the number elements from which the predictions are made.
		
		buckets is the resolution of the false positive axis. It defines the number
		of divisions. A value of 0 means to have as many buckets as input values.

		important can be replaced with an integer, to change the number of elements
		to be evaluated.
		'''

		#allow the user to change how many elements are evaluated.
		self.setImportant(important)

		#sort descending
		positions = sorted(positions)

		#keep all positions
		allPositions = positions[:]

		#if we are supposed to keep only a couple of instances
		if self.important>0:
			#keep the first positions
			positions = [x for e,x in enumerate(allPositions) if (x-e)<self.important]
			#list total elements available
			totalElements = self.important+len(positions)
		
		if len(positions) == 0:
			if self.__verbose__:
				print 'ROC: 0, no results found in the first',self.important,'positions'
			return self.getInfLimit()

		falseElements=totalElements-len(positions)

		#define the amount of false positives per bucket
		if buckets > 0:
			if buckets > falseElements:
				#max amount of buckets
				buckets = falseElements

			bucketSize = falseElements / float(buckets)
			if self.__debug__:
				print 'bucket size:',bucketSize
		else:
			buckets = falseElements
			bucketSize = 1.0

		#percent of true positive increase
		yDelta = 1.0/len(allPositions)
		#percent of false positive increase
		xDelta = 1.0/(buckets)
		
		#histogram of the true labels
		trueHistogram = [0 for x in range(buckets)]

		for e,pos in enumerate(positions):
			#put the true label in the proper bucket. 
			#Skip the number of true labels as they don't count towards the x axis of the plot
			bucket=(int)((pos-e)/bucketSize)
			try:
				#skip if the true label is the last predicted element
				trueHistogram[bucket]+=1
			except:
				print e,pos,'cannot be put in a bucket'
				#pass
		cumHistogram = np.cumsum(trueHistogram)

		cumROC = cumHistogram.sum()*xDelta*yDelta

		if self.__verbose__:
			print 'AUC:',cumROC,'for',buckets,'buckets, from',totalElements,'elements and',len(allPositions),'true labels'

		return cumROC

	def improves(self, a, b):
		'''
		Returns whether a improves b considering this measure
		'''

		return a>b

	def setImportant(self,important):
		'''
		Sets how many elements should be considered for the ROC curve
		'''
		if important:
			self.important = important

	def getInfLimit(self):
		'''
		Returns the inferior limit of the measure
		'''

		return 0.0

	def getSupLimit(self):
		'''
		Returns the superior limit of this measure
		'''

		return 1.0

	def randomGuess(self,totalElements):
		'''
		Random guess of this measure.

		totalElements is the number elements from which the predictions are made.
		'''

		#average ROC = 0.5, scaled to the first x elements
		return (self.important*0.5)/totalElements

class VectorMean:
	def __init__(self, verbose=False):
		self.__verbose__=verbose

	def evaluate(self, positions, totalElements=0):
		'''
		Obtains the mean of the positions in the vector

		positions is a list with the positions of the predicted elements.

		totalElements is the number elements from which the predictions are made.
		'''
		#positions is a 0 indexed vector, we must add one per prediction

		out = float(sum(positions)+len(positions))/len(positions)
		if self.__verbose__:
			print 'Average position of targets', out
		return float(sum(positions)+len(positions))/len(positions)

	def improves(self, a, b):
		'''
		Returns whether a improves b considering this measure
		'''

		return a<b

	def getInfLimit(self):
		'''
		Returns the inferior limit of the measure
		'''

		return 100000

	def getSupLimit(self):
		'''
		Returns the superior limit of this measure
		'''

		return 0

	def randomGuess(self,totalElements):
		'''
		Random guess of this measure.

		totalElements is the number elements from which the predictions are made.
		'''

		#average position would be the half of the total elements
		return 0.5*totalElements

class L1oRecall:
	'''
	Evaluates whether the left out element comes in the top x elements.
	'''

	def __init__(self, verbose=False):
		self.__verbose__=verbose

	def evaluate(self, position, top=200):
		'''
		Returns 1 if position is within the top x elements.

		if the position is negative, we assume that the element
		cannot be found in the set.
		'''

		#position not found within recall set
		if position < 0:
			return 0

		if position<top:
			return 1
		else:
			return 0

	def improves(self, a, b):
		'''
		Returns whether a improves b considering this measure
		'''

		return a<b

	def getInfLimit(self):
		'''
		Returns the inferior limit of the measure
		'''

		return 0.0

	def getSupLimit(self):
		'''
		Returns the superior limit of this measure
		'''

		return 1.0

	def randomGuess(self,top,totalElements):
		'''
		Random guess the average results for this measure after an infinite
		amount of trials, given an uniform distribution for the position of
		the correct label within the vector.

		top are how many elements are being analysed

		totalElements is the number elements from which the predictions are made.
		'''

		#probability of finding the correct within the first x elements
		return float(top)/totalElements