"""
    Cardigan - Disease Gene Prediction Method
    Copyright (C) 2018  Juan Caceres

    Cardigan is licensed under a Creative Commons Attribution-NonCommercial 4.0 
    International License.  If you did not get a copy of the license with
    the source code, you can get a copy at:
    
    https://creativecommons.org/licenses/by-nc/4.0/legalcode

"""
__author__  = "Juan Caceres"
__email__   = "Juan.CaceresSilva.2014@live.rhul.ac.uk"
__copyright__ = "Copyright (C) 2018 Juan Caceres"
__license__ = "CC BY-NC"
__version__ = "1.0"

from collections import defaultdict
from tsvParser import Parser
import re
import numpy as np

class MorbidMap:
	'''
	Class to handle diseases (MIM) and the genes related to those diseases.
	'''
	def __init__(self, filename, mim2gene=None, oldMorbidmap=None, columns=False, 
		separator='\t', verbose=True, reverseDict=False, names=False):
		'''
		-filename should be the path to a tsv morbidmap file

		-Columns is a list of the columns the file contains. 
			The file must contain at least the following columns:
				'phenotype','mim_number'
		-mim2gene
		'''

		self.__verbose__ = verbose
		
		self.filename = filename
		if columns:
			self.__columns__ = columns
		else:
			# Phenotype	Gene Symbols	MIM Number	Cyto Location
			self.__columns__ = ['phenotype','gene_symbols','mim_number','cyto_location']

		#Mim2Gene object, translates mims and gene symbols (i.e. entrez identifiers)
		self.mim2gene = mim2gene

		#start tsvParser Parser object on the given file
		self.__parser = Parser(self.filename, self.__columns__, separator)
		#save the dictionary of mim -> 
		self.map = self.parse(reverseDict, names)		

		

		#if an old morbidmap is included, calculate the differences and store them
		if oldMorbidmap:
			self.geneAdditions = self.calculateGeneAdditions(oldMorbidmap)
			self.diseaseAdditions = self.calculateDiseaseAdditions(oldMorbidmap)

	def parse(self, reverseParse = False, names = False):
		'''
		Makes a dictionary of disease -> disease protein set
		
		disease is a MIM
		
		protein is a entrez id (if mim2gene is provided) or a MIM otherwise
		
		names is a boolean that allows the creation of a mim to disease name dictionary
		accessible through self.namesMap
		'''
		dictionary = defaultdict(set)
		allMims = set()
		translatedMims = set()
		
		if reverseParse:
			reverseDict = defaultdict(set)
		
		if names:
			namesDict = defaultdict(set)
		association_count=0
		lines=0
		hasMimRegex = re.compile('[0-9]{6}')
		for parsedLine in self.__parser:
			lines+=1

			#grab the disease mim is possible, otherwise use the whole disease name
			#there is no filtering currently
			if len(hasMimRegex.findall(parsedLine.phenotype))>0:              
				diseaseMim=hasMimRegex.findall(parsedLine.phenotype)[0]
			else:
				diseaseMim=parsedLine.mim_number
				#continue

			# get all MIMs that appear to be genes
			allMims |= {parsedLine.mim_number}

			#if the mim translation dicitonary is available
			if self.mim2gene:
				#if the mim can be translated
				if self.mim2gene.getEntrez(parsedLine.mim_number):
					#translate the mim to an identifier
					geneLabel = self.mim2gene.getEntrez(parsedLine.mim_number)
					#add this mim as a succesfully translated mim
					translatedMims |= {parsedLine.mim_number}
					#count the association
					association_count+=1
				else:
					#print proteinMim
					#do not add the mim to the dictionary
					continue
			else:
				#if no translation dictionary is available
				geneLabel = parsedLine.mim_number
				#count the association
				association_count+=1
				
			dictionary[diseaseMim] |= {geneLabel}
			if reverseParse:
				reverseDict[geneLabel] |= {(diseaseMim,parsedLine.phenotype)}
				
			if names:
				namesDict[diseaseMim] = parsedLine.phenotype
		
		if reverseParse:
			self.reverseMap = reverseDict
		
		if names:
			self.namesMap = namesDict

		if self.__verbose__:
			print 'Parsed',len(dictionary),'diseases and',len(allMims),'genes,',\
					len(allMims)-len(translatedMims),'genes could not be translated',\
					'from',association_count,'associations and',lines,'lines'

		return dictionary

	def getDiseasesForGene(self, gene):
		'''
		Returns the diseases that contain a gene
		'''
		try:
			return self.reverseMap[gene]
		except:
			return 'No gene name map loaded, try myMorbidmap = (reverseDict = True)'
	
	def getDiseaseName(self, mim):
		'''
		Returns the phenotype from the mim
		'''
		try:
			return self.namesMap[mim]
		except:
			return 'No disease name map loaded, try myMorbidmap = (names = True)'
	
	def morbidmapTagParser(self, filename,separator='\t'):
		'''
		Makes a dictionary of disease -> words
		disease is a MIM
		words are the words in the name of the disease
		'''
		dictionary = defaultdict(set)
		allProteins = set()
		translatedProteins = set()
		hasMimRegex = re.compile('[0-9]{6}')
		for parsedLine in self.__parser:
			#get disease mim from the disease name if possible
			if len(hasMimRegex.findall(parsedLine.phenotype))>0:              
				diseaseMim=hasMimRegex.findall(parsedLine.phenotype)[0]
			else:
				#the disease mim is in the column
				diseaseMim=parsedLine.mim_number
				#continue


			#get lowercase 
			phenotype = parsedLine.phenotype.replace('-',' ').lower()
			phenotype = phenotype.replace(',','')

			for tag in phenotype.strip().split(" "):
				dictionary[tag]|= {mim}

		return dictionary

	def calculateGeneAdditions(self, oldMorbidmap):
		'''
		Makes a dictionary mim -> genes

		Where mim is each entry of the old morbidmap, and genes are the genes this
		morbidmap includes, but the old does not.
		'''
		#return dictionary
		dictionary=defaultdict(set)

		#get possible mims for additoins
		persistentMims =  set(self.map.keys()) & set(oldMorbidmap.map.keys())

		#for every entry of the old map
		for mim in persistentMims:
			#get the new mapped genes
			genes = self.map[mim]-oldMorbidmap.map[mim]

			#only annotate if there are any new genes
			if len(genes)>0:
				dictionary[mim]=genes

		if self.__verbose__:
			if len(dictionary)>0:
				print 'Gene Additions:'
				print 'From',len(oldMorbidmap.map),'diseases,',len(dictionary),'have new genes'
				print 'Found',sum([len(genes) for genes in dictionary.values()]),\
					'cumulative new genes for the diseases and',\
					sum([len(genes-set.union(*oldMorbidmap.map.values())) for genes in dictionary.values()]),\
					'novel genes from a total of',len(set.union(*dictionary.values())),'different new genes, and',\
					len(set.union(*dictionary.values())-set.union(*oldMorbidmap.map.values())),\
					'novel genes'
			else:
				print 'No Gene Additions'

		return dictionary


	def calculateDiseaseAdditions(self, oldMorbidmap):
		'''
		Makes a dictionary mim -> genes

		where mim is included in this morbidmap but not in the old morbidmap
		'''
		#return dictionary
		dictionary=defaultdict(set)

		count = 0
		#get the keys not included in the old morbidmap
		newMims =  set(self.map.keys())-set(oldMorbidmap.map.keys())
		for mim in newMims:
			#add the genes
			dictionary[mim] = self.map[mim]
			if len(dictionary[mim])>0:
				count += 1

		if self.__verbose__:
			if len(dictionary)>0:
				print 'Disease Additions:'
				print 'From',len(self.map),'diseases,',len(dictionary),'were orphan'
				print 'Found',sum([len(genes) for genes in dictionary.values()]),\
					'cumulative new genes for the diseases and',\
					sum([len(genes-set.union(*oldMorbidmap.map.values())) for genes in dictionary.values()]),\
					'novel genes, from a total of',\
					len(set.union(*dictionary.values())),'different new genes, and',\
					len(set.union(*dictionary.values())-set.union(*oldMorbidmap.map.values())),\
					'novel genes'
			else:
				print 'No Disease Additions'
		return dictionary

	def getGenes(self, mim):
		'''
		Retrieves the genes referred by the MIM
		'''
		if isinstance(mim, (tuple, list, set)):
			#try to use mim as an iterable
			out = set()
			for m in mim:
				if m in self.map:
					out |= self.map[m]
			return out
		else:
			#check to avoid adding empty entries to the map
			if mim in self.map:
				return self.map[mim]
			else:
				return set()

	def getDiseases(self):
		'''
		Retrieves all the annotated disease MIMs
		'''

		return self.map.keys()

	def saveAsDiamond(self, outfile = 'morbidmapDiamond.txt', names = False):
		'''
		Saves the morbidmap as the Diamond input
		'''

		with open(outfile, 'wb') as f:
			f.write('#disease\tgenes\n')

			for disease, genes in self.map.iteritems():

				if names:
					name = self.namesMap[disease]
				else:
					name = disease

				f.write(name+'\t'+'/'.join(genes)+'\n')

		if self.__verbose__:
			print 'Created',outfile,'with',len(self.map),'diseases'

	def saveAsProdige(self, genesInPPIFile = '9670GeneId.txt', outfile = 'morbidmapProdige.txt',
		diseasesOutfile = 'diseasesForProdige.txt', associationPairsOutfile = 'associationPairsProdige.txt',
		diseasesToInclude = False):
		'''
		Saves the morbidmap as the Prodige input associations file. 

		The output is a two column file with a MIM number in the first column
		and a gene position in the second column. The gene position must be
		the line number of the gene name inside the genesInPPIFile.

		genesInPPIFile is a text file with one gene name per line, as used by
		Prodige. See 9670GeneId.txt as an example if needed.

		outfile is the name of the output file

		diseasesToInclude is an optional list to specify the list of disease
		MIMs to use in the prodige morbidmap (possibly including diseases with
		no known genes from the vanDriel matrix)

		'''

		geneList = []

		#read each line into an array
		with open(genesInPPIFile, 'r') as f:
			for line in f:
				line = line.strip()
				geneList.append(line)

		foundDiseases = []
		count = 0
		with open(outfile, 'wb') as f:
			#for each disease
			for disease in sorted(self.map.keys()):
				genes = self.map[disease]
				ban = False
				#for every gene
				for gene in genes:
					#if the gene can be mapped to the PPI
					if gene in geneList:
						ban = True
						#write the disease-gene association
						f.write(disease+'\t'+str(geneList.index(gene)+1)+'\n')

				if ban:
					foundDiseases.append(disease)
					count += 1
		if self.__verbose__:
			print 'Created',outfile,'with',count,'diseases'	

		if diseasesToInclude:
			#add the diseases to include to our output set
			foundDiseases = sorted(set(foundDiseases)|set(diseasesToInclude))

		with open(diseasesOutfile, 'wb') as f:
			f.write('\n'.join(foundDiseases))
		if self.__verbose__:
			print 'Created',diseasesOutfile,'with',len(foundDiseases),'diseases'

		count = 0

		with open(associationPairsOutfile, 'wb') as f:
			#for each disease
			#diseaseCount = 0
			for diseaseCount, disease in enumerate(foundDiseases):
				genes = self.map[disease]
				ban = False
				#for every gene
				for gene in genes:
					#if the gene can be mapped to the PPI
					if gene in geneList:
						#ban = True
						#write the disease-gene association
						f.write(str(diseaseCount*len(geneList)+geneList.index(gene)+1)+'\n')
						count += 1
				#if ban:
				#	diseaseCount += 1
		if self.__verbose__:
			print 'Created',associationPairsOutfile,'with',count,'associations, and',len(foundDiseases),'diseases'
		


	def saveDiseaseNameToDiseaseMim(self, outfile = 'diseaseNameToDiseaseMim.txt'):
		'''
		Makes a two column file, the first column is the disease name, while
		the second column is the disease MIM. This file will produce something
		alike the Diamond "disease to MIM" file
		'''

		#try:

		with open(outfile, 'wb') as f:
			f.write('#Diseases\tMIMS\n')

			for mim, name in self.namesMap.iteritems():
				f.write(name+'\t'+mim+'\n')

		if self.__verbose__:
			print 'Created',outfile,'with',len(self.namesMap),'diseases'

#		except NameError:
			#print 'Name map is probably missing, call self.parse with names=True'
			#_ = self.parse(names=True)
			#self.saveDiseaseNameToDiseaseMim(outfile)

	def toArray(self):
		'''
		Convert morbidmap to a numpy array
		'''

		arrayRows = sorted(self.getDiseases())
		arrayCols = sorted(self.getGenes(arrayRows))

		mmArray = np.zeros((len(arrayRows),len(arrayCols)))
		for i,disease in enumerate(arrayRows):
			genes = self.getGenes(disease)
			for gene in genes:
				mmArray[i][arrayCols.index(gene)]=1

		self.__array__ = mmArray
		self.__arrayRows__ = arrayRows
		self.__arrayCols__ = arrayCols



class Mim2Gene:
	'''
	Class to handle the identification of genes by their MIM number
	'''
	def __init__(self, filename, columns=False, separator='\t', verbose=False):
		'''
		- filename: is the file to be opened

		- columns: is an optional parameter defining the columns from the mim2gene file.
		It must include 'mim_number', 'entry_type', and at least one of 'entrez_id', 'symbol'
		and/or 'ensembl_id'.

		- separator: is the column separator

		'''
		self.filename = filename

		if columns:
			self.__columns__ = columns
		else:
			# MIM Number	MIM Entry Type (see FAQ 1.3 at http://omim.org/help/faq)	Entrez Gene ID (NCBI)	Approved Gene Symbol (HGNC)	Ensembl Gene ID (Ensembl)
			self.__columns__ = ['mim_number', 'entry_type', 'entrez_id', 'symbol', 'ensembl_id']
		try:
			#start tsvParser Parser object on the given file
			self.__parser = Parser(self.filename, self.__columns__)
			#parse the file
			self.translation = self.parse(verbose)
		except:
			self.__columns__ = ['mim_number', 'entry_type', 'entrez_id', 'symbol']
			self.__parser = Parser(self.filename, self.__columns__)
			self.translation = self.parse(verbose)

	def parse(self, verbose = True):
		'''
		Makes a dictionary of mim -> entrez,symbol
		mim is a 6 digit number
		in the tuple (entrez, symbol)
			entrez is a number
			symbol is a string
		'''
		dicitonary = dict()

		genes = 0
		phenotypes = 0

		for parsedLine in self.__parser:
			fields = dict()
			#entry type
			fields['type'] = parsedLine.entry_type
			if 'entrez_id' in self.__columns__:
				#entrez identifier
				fields['entrez'] = parsedLine.entrez_id
			if 'symbol' in self.__columns__:
				#official HGNC symbol
				fields['symbol'] = parsedLine.symbol
			if 'ensembl_id' in self.__columns__:
				#ensembl identifier
				fields['ensembl'] = parsedLine.ensembl_id

			if parsedLine.entry_type == 'gene':
				genes += 1
			if parsedLine.entry_type == 'gene/phenotype':
				genes += 1
				phenotypes += 1
			if parsedLine.entry_type == 'phenotype':
				phenotypes += 1

			#use the mim number as the key
			dicitonary[parsedLine.mim_number] = fields

		if verbose:
			print 'Parsed',genes,'genes and',phenotypes,'phenotypes'

		return dicitonary

	def getEntrez(self, mim, verbose = False):
		'''
		Retrieves the entrez from a mim. 
		Returns False if there is an error.
		'''
		if mim in self.translation:
			if self.translation[mim]['entrez']:
				return self.translation[mim]['entrez']
			else:
				if verbose:
					print 'No entrez associated for MIM',mim
				return False

		if verbose:
			print 'MIM not found in the dictionary'

		return False

	def getEnsembl(self, mim, verbose = False):
		'''
		Retrieves the ensembl from a mim. 
		Returns False if there is an error.
		'''
		if mim in self.translation:
			if self.translation[mim]['ensembl']:
				return self.translation[mim]['ensembl']
			else:
				if verbose:
					print 'No ensembl associated for MIM',mim
				return False

		if verbose:
			print 'MIM not found in the dictionary'

		return False

	def getSymbol(self, mim, verbose = False):
		'''
		Retrieves the symbol from a mim. 
		Returns False if there is an error.
		'''
		if mim in self.translation:
			if self.translation[mim]['symbol']:
				return self.translation[mim]['symbol']
			else:
				if verbose:
					print 'No symbol associated for MIM',mim
				return False

		if verbose:
			print 'MIM not found in the dictionary'

		return False

class DiseaseModules:
	'''
	Class to handle the DIAMOnD disease modules
	'''
	def __init__(self, diseaseGeneFile, diseaseMimFile, verbose=False):
		'''
		- diseaseGeneFile: is a TSV file containing disease names and associated genes
		 (i.e. the 201504-18_PLoS-DIAMOnD-S2-Data.txt file from DIAMOnD)

		- diseaseMimFile: is a TSV file containing disease names and associated MIM ids
		 (i.e. the diseasesToMIM.txt file provided with Cardigan)
		'''

		self.d2g = self.parse(diseaseGeneFile)
		self.d2m = self.parse(diseaseMimFile)

	def parse(self, filename):

		out = defaultdict(list)

		with open(filename,'r') as f:
			for line in f:
				if line[0]='#':
					continue

				line = line.split()
				name = ' '.join(line[:-1])
				values = line[-1].split('/')

				out[name] = values

		return out

	def getGenes(self, diseases):
		'''
		Retrieves the genes referred by the Disease

		-diseases can be a list or a string
		'''
		return self.__getDictValues(diseases,self.d2g)


	def getMims(self, diseases):
		'''
		Retrieves the genes referred by the Disease

		-diseases can be a list or a string
		'''
		return self.__getDictValues(diseases,self.d2m)

	def __getDictValues(self, identifiers, dictionary):
		'''
		Retrieves identifiers from the dictionary
		'''
		if isinstance(identifiers, (tuple, list, set)):
			#try to use identifiers as an iterable
			out = set()
			for m in identifiers:
				if m in dictionary:
					out |= dictionary[m]
			return out
		else:
			#check to avoid adding empty entries to the map
			if identifiers in dictionary:
				return dictionary[identifiers]
			else:
				return set()
