"""
    Cardigan - Disease Gene Prediction Method
    Copyright (C) 2018  Juan Caceres

    Cardigan is licensed under a Creative Commons Attribution-NonCommercial 4.0 
    International License.  If you did not get a copy of the license with
    the source code, you can get a copy at:
    
    https://creativecommons.org/licenses/by-nc/4.0/legalcode

"""
__author__  = "Juan Caceres"
__email__   = "Juan.CaceresSilva.2014@live.rhul.ac.uk"
__copyright__ = "Copyright (C) 2018 Juan Caceres"
__license__ = "CC BY-NC"
__version__ = "1.0"

from collections import defaultdict
from tsvParser import Parser
from datetime import datetime
import numpy as np
from scipy import sparse

class PPI:
	def __init__(self, filename, columns='hippie', ppiName='entrez_pair', mainIdentifier=None, separator='\t',
		synSeparator='|', translateFile=None, translateCols=['source','target'], verbose=True,simplify=True):
		'''
		-filename should be the path to a tsv file with the PPI data

		-Columns is a list of the columns the file contains. 
			The file must contain at least the following columns:
				'entrez_a','entrez_b'
			it can also be one of the following strings, which will be translated to the proper lists:
				'biogrid', 'hippie', 'funcoup'
		-mainIdentifier must be one of the column names or 'symbol', used as gene identifier.
		i.e. if column names are 'entrez_a','entrez_b', the mainIdentifier should be 'entrez'

		-translateFile is an external file to try to translate the identifier from the ppi
		to another identifier (translates if mainIdentifier is not 'symbol')

		-translateCols are the columns of the translation file, it must contain the columns 'source'
		and 'target', which are the only used columns from the file.

		-simplify is a boolean to remove self links and multi links.

		'''

		self.verbose=verbose

		self.filename = filename

		#just asthetics
		self.name = ppiName

		#Detect missing identifiers
		self.__nullIdentifiers__ = ['','-',' ',None]

		#Remove self links and multigraph properties
		self.__simplify__ = simplify

		#general fallback identifier, in case the main identifier fails
		self.__fallbackIdentifier__ = 'symbol'

		self.__columns__ = columns
		if columns.lower() == 'biogrid':
			##BioGRID Interaction ID	Entrez Gene Interactor A	Entrez Gene Interactor B	BioGRID ID Interactor A	BioGRID ID Interactor B	Systematic Name Interactor A	Systematic Name Interactor B	Official Symbol Interactor A	Official Symbol Interactor B	Synonyms Interactor A	Synonyms Interactor B	Experimental System	Experimental System Type	Author	Pubmed ID	Organism Interactor A	Organism Interactor B	Throughput	Score	Modification	Phenotypes	Qualifications	Tags	Source Database
			self.__columns__ = ['biogrid_id','entrez_a','entrez_b','biogrid_id_a','biogrid_id_b','name_a',
			'name_b','symbol_a','symbol_b','synonyms_a','synonyms_b','experimental_system',
			'experimental_system_type','author','pubmed_id','organism_a','organism_b','throughput','score',
			'modification','phenotypes','qualifications','tags','source_db']
			self.mainIdentifier = 'entrez'
			self.__fallbackIdentifier__ = 'biogrid_id'
		if columns.lower() == 'hippie':
			self.__columns__ = ['symbol_a','entrez_a','symbol_b','entrez_b','weight','references']
			self.mainIdentifier = 'entrez'
		if columns.lower() == 'hprd':
			self.__columns__ = ['symbol_a','hprd_a','refseq_a','symbol_b','hprd_b','refseq_b','experiment_type','pubmed_id']
			self.mainIdentifier = 'symbol'
			self.__fallbackIdentifier__='refseq'
		if columns.lower() == 'entrez_pair':
			self.__columns__ = ['entrez_a','entrez_b']
			self.mainIdentifier = 'entrez'
			self.__fallbackIdentifier__='entrez'
		if columns.lower() == 'entrez_pair_weighted':
			self.__columns__ = ['entrez_a','entrez_b','weight']
			self.mainIdentifier = 'entrez'
			self.__fallbackIdentifier__='entrez'
		if columns.lower() == 'diamondnet':
			self.__columns__ = ['entrez_a','entrez_b','symbol_a','symbol_b','sources']
			self.mainIdentifier = 'entrez'
		if columns.lower() == 'funcoup':
			#fbs = final bayesian score
			#llr = log likelihood ratio
			self.__columns__ = ['weight', 'fbs_max', 'ensembl_a', 'ensembl_b', 'fbs_ppi', 'fbs_metabolic',
			'fbs_signaling', 'fbs_complex', 'llr_mex', 'llr_ppi', 'llr_scl', 'llr_php', 'llr_mir', 'llr_tfb',
			'llr_pex', 'llr_gin', 'llr_dom', 'llr_hsa', 'llr_mmu', 'llr_rno', 'llr_cfa', 'llr_gga', 'llr_dre',
			'llr_cin', 'llr_dme', 'llr_cel', 'llr_sce', 'llr_ath', 'max_category']
			self.mainIdentifier = 'ensembl'
			self.__fallbackIdentifier__='ensembl'

		#Which column of the PPI file is used as the main identifier (overrides the expected main identifier)
		if mainIdentifier:
			self.mainIdentifier = mainIdentifier

		if translateFile:
			self.createTranslationDict(translateFile, translateCols)

		#start tsvParser Parser object on the given file
		self.__parser = Parser(self.filename, self.__columns__)
		
		#load data from the file
		self.parse(synSeparator)
		self.__createEdgeIndices__()
		
			
	def createTranslationDict(self, translateFile, translateCols):
		'''Creates an identifier translation dictionary.

		translateFile must be a tsv file with translateCols as columns.

		translateCols is a list of strings, it must include 'from' and 'to'
		'''
		self.__translate__ = dict()

		parser = Parser(translateFile, translateCols)

		for parsedLine in parser:
			if len(parsedLine.source)>0 and len(parsedLine.target)>0:
				self.__translate__[parsedLine.source]=parsedLine.target

		if self.verbose:
			print 'Found translations for up to',len(self.__translate__),'genes'

	def getGenesFromLine(self,parsedLine):
		'''
		Returns a pair of genes from the parsed line

		Tries to retrieve the main identifier, if not possible, it uses
		the fallback identifier before failing
		'''
		geneA = self.__getGeneFromLine(parsedLine,'_a')
		geneB = self.__getGeneFromLine(parsedLine,'_b')

		return geneA,geneB

	def __getGeneFromLine(self,parsedLine,which='_a'):
		'''
		Returns a single gene from the parsed line

		which should be either '_a' or '_b'
		'''
		gene=getattr(parsedLine,self.mainIdentifier+which)
		if gene in self.__nullIdentifiers__:
			gene=getattr(parsedLine,self.__fallbackIdentifier__+which)
		if gene in self.__nullIdentifiers__:
			gene = None
		else:
			try:
				gene=self.__translate__[gene]
			except:
				pass

		return gene

	def parse(self, synSeparator):
		'''
		returns
		-list of vertices
		-list of edges
		-dictionary of protein symbols (synonyms) to the official protein symbol,
			used included in the vertices/edges
		'''
		#every vertex parsed
		vertices = set()
		#every edge found (1 per line)
		edges = []
		edgeGraph = dict()
		#weights
		weights = False
		#if using symbols it will point every synonym to the official symbol
		identifierDict = defaultdict(set)
		#errors of official symbol referencing
		errors=set()


		for parsedLine in self.__parser:

			geneA,geneB = self.getGenesFromLine(parsedLine)

			if not geneA or not geneB:
				continue

			vertices.add(geneA)
			vertices.add(geneB)

			if self.__simplify__ and geneA == geneB:
				continue

			try:
				if edgeGraph[(geneA,geneB)]<float(parsedLine.weight):
					edgeGraph[(geneA,geneB)]=float(parsedLine.weight)
					edgeGraph[(geneB,geneA)]=float(parsedLine.weight)
				weights = True
			except:
				try:
					edgeGraph[(geneA,geneB)]=float(parsedLine.weight)
					edgeGraph[(geneB,geneA)]=float(parsedLine.weight)
					weights = True
				except:
					edgeGraph[(geneA,geneB)]=1
					edgeGraph[(geneB,geneA)]=1

			#if self.mainIdentifier == 'symbol' and not hasattr(self, '__translate__'):			
			if self.mainIdentifier == 'symbol':			
				try:
					for symbol,synonyms in [(parsedLine.symbol_a,parsedLine.synonyms_a),
							(parsedLine.symbol_b,parsedLine.synonyms_b)]:

						#skip null symbols
						if symbol in self.__nullIdentifiers__:
							continue
						#map every symbol to the official symbol
						identifierDict[symbol] = {symbol}
						#will only work if there is any synonyms present
						synonyms = synonyms.split(synSeparator)

						for syn in synonyms:
							#skip the null synonyms
							if syn in self.__nullIdentifiers__:
								continue
							identifierDict[syn] |= {symbol}
				except AttributeError:
					#if the PPI file has no synonyms
					pass

		#fix identifier dict possible errors
		for key,val in identifierDict.items():
			if len(val)>1:
				#the symbol is a synonym of more than one official symbol
				if key in val:
					#if the symbol itself is an official symbol, just don't use it as a synonym
					identifierDict[key]={key}
				else:
					errors.add(key)
					if debug:
						print key,val
			#use the first match as the official symbol
			identifierDict[key]=list(val)[0]

		if self.verbose and len(identifierDict)>0:
			print 'Overlapping synonyms:',len(errors)

		vertices = list(vertices)

		self.vertices = vertices
		self.edges = edgeGraph.keys()
		if weights:
			self.weights = edgeGraph.values()
		else:
			self.weights = []
		self.identifierDict = identifierDict
		self.errors = errors

		return vertices,edges,weights,identifierDict,errors

	def __createEdgeIndices__(self):
		'''
		create edges pointing to the vertex index instead of vertex name
		'''
		ids = {x:i for i,x in enumerate(self.vertices)}

		self.edgeIndices=[]
		weights=[]
		for i,edge in enumerate(self.edges):
			if ids[edge[0]]<ids[edge[1]]:
				self.edgeIndices.append((ids[edge[0]],ids[edge[1]]))
				if self.weights:
					weights.append(self.weights[i])

		if self.weights:
			self.weights=weights


		#return self.edgeIndices

	def __findConnectedComponents__(self):
		'''
		Returns a list of sets with the connected components in the graph
		'''

		clusters = [{x} for x in self.vertices]
		cluster_index = {x:i for i,x in enumerate(self.vertices)}

		#join all pairs of clusters based on the edges
		for edge in self.edges:
			#find the index of the cluster on each node of the edge
			ix=[cluster_index[edge[0]],cluster_index[edge[1]]]
			#if the nodes belong to different clusters
			if ix[0] != ix[1]:
				#join them on the cluster 0
			 	clusters[ix[0]] = clusters[ix[0]] | clusters[ix[1]]
			 	for x in clusters[ix[1]]:
			 		#change the cluster index of the nodes in cluster 1 to cluster 0
			 		cluster_index[x]=ix[0]
			 	#delete the obsolete cluster 1
			 	clusters[ix[1]] = []

		self.__clusters__ = [x for x in clusters if x]


	def getGeneIndices(self, geneList):
		'''
		Retrieves the index in the graph of the genes
		'''

		vertices = list()
		for gene in geneList:
			vertices.append(self.getGeneIndex(gene))

		return vertices

	def getGeneIndex(self, gene):
		'''
		Retrieves the index in the graph of the gene
		'''

		try:
			return self.vertices.index(gene)
		except:
			return False


	def edgesToSparseMatrix(self):
		'''
		Make a sparse adjacency matrix from a list of edges
		'''
		xs, ys = map(np.array, zip(*self.edgeIndices))
		xs, ys = np.hstack((xs, ys)).T, np.hstack((ys, xs)).T
		if self.weights:
			elements = np.array(self.weights+self.weights)
		else:
			elements = np.ones(xs.shape)

		return sparse.coo_matrix((elements, (xs, ys))) 

	def getGenesInGraph(self, geneList):
		'''
		Obtains the subset of genes that are mapped in the graph.

		geneList must be a list of gene identifiers, as they are named in the PPI.
		'''

		return list(set(self.vertices) & set(geneList))

	def _getConnectedGenes(self, geneList='all', ccSize=1, hardConnectivity=True):
		'''		
		Obtains the subset of genes from which you can transverse to at least
		'ccSize' other genes.

		geneList must be a list of gene identifiers, as they are named in the PPI.

		ccSize is the minimum number of extra reachable genes.

		hardConnectivity requires for each returned gene to be the source of
		the connection. When false, at lest 'ccSize' elements should be reachable
		using all the genes from the geneList.
		'''

		if not geneList:
			return geneList

		if geneList == 'all':
			genes = self.vertices
		else:
			genes = self.getGenesInGraph(geneList)

		if not genes:
			return []

		try:
			len(self.__clusters__)
		except:
			self.__findConnectedComponents__()
	
		#all the nodes that are in big enough connected components
		reachable = set()

		for gene in genes:
			#see if the cluster containing the gene was added
			if gene in reachable:
				continue

			#search for the cluster which contains the current gene
			for cluster in self.__clusters__:
				if gene in cluster:
					#when hardConnectivity is required
					if hardConnectivity & len(cluster) > ccSize:
						#we add only clusters bigger than the ccSize
						reachable |= set(cluster)
					#otherwise
					if not hardConnectivity:
						#we add clusters of any size
						reachable |= set(cluster)							
					#either way, we have already found the cluster for the current gene
					continue
				
			
		#return the genes we could find
		return [x for x in genes if x in reachable]