"""
    Cardigan - Disease Gene Prediction Method
    Copyright (C) 2018  Juan Caceres

    Cardigan is licensed under a Creative Commons Attribution-NonCommercial 4.0 
    International License.  If you did not get a copy of the license with
    the source code, you can get a copy at:
    
    https://creativecommons.org/licenses/by-nc/4.0/legalcode

"""
__author__  = "Juan Caceres"
__email__   = "Juan.CaceresSilva.2014@live.rhul.ac.uk"
__copyright__ = "Copyright (C) 2018 Juan Caceres"
__license__ = "CC BY-NC"
__version__ = "1.0"

from collections import defaultdict
from tsvParser import Parser
import numpy as np
from scipy import sparse
import scipy.sparse.linalg as la
from scipy.special import expit
from datetime import datetime
import sys
import abc


import Measures

'''
To run a dataset you want to have the following things
- The affinity matrix
- The diffusion matrix
- Run the prediction

Example:
```python
W=calculateAffinityMatrix(X)
S=calculateDiffusionMatrix(W)
F=solveRegularisedSystem(S)
```
where `X` is a matrix with the points you want to use.

`W` can otherwise be obtained as the graph's affinity matrix (either weighted
or unweighted). This matrix can be a numpy np.array or scipy sparse in the
current implementation.

Review `solveRegularisedSystem` to see how to feed the initial labels for the
system '''

class Diffusion:
	def __init__(self, ppi, morbidMap, diseaseSimilarities, verbose=True):
		'''
		ppi is a PPIUtilities.PPI object
		morbidMap is an OMIMUtilities.MorbidMap object
		mim2gene is an OMIMUtilities.Mim2Gene object
		'''

		self.ppi = ppi
		self.mm = morbidMap
		self.ds = diseaseSimilarities
		self.__verbose__ = verbose

		self.measure = 'roc'


		#seeds for the diffusion
		self.seeds = dict()
		self.seeds['genes'] = list()
		self.seeds['diseases'] = list()
		self.targets = list()

		#load the affinity matrix
		self.W = self.ppi.edgesToSparseMatrix()

		#variables available after diffusion
		self.lastGeneWeights = None
		self.lastDiffusionVector = None
		self.predictionResult = None

	def __calculateDiffusionMatrix(self, weightMatrix = None):
		'''
		Creates a Diffusion matrix out of the affinity matrix W.

		The affinity matrix may be overriden if given as a parameter.
		'''

		#override weight matrix if needed
		if weightMatrix:
			self.W = weightMatrix

		#build D
		sums=self.W.sum(1) #sum every column per row

		#in case there are unconnected parts in the matrix
		indNonZeros = np.where(sums != 0)
		diagonalValues = np.zeros(sums.shape)

		diagonalValues[indNonZeros] = 1.0/np.sqrt(sums[indNonZeros])
		D = sparse.spdiags(diagonalValues.T,0,diagonalValues.shape[0],diagonalValues.shape[0])

		#build S
		S=D*self.W*D

		return S


	def __prepareGraphToDiffuse(self, verbose = False):
		'''
		DEPRECATED

		Helper function to create missing variables in order to diffuse
		'''

		if not hasattr(self, 'W'):
			#build the affinity matrix - as and adjacency matrix
			self.W = self.ppi.edgesToSparseMatrix()

		if not hasattr(self, 'S'):
			#build the diffusion matrix
			self.S = self.__calculateDiffusionMatrix()

	def setDiffusionGenes(self, geneList, verbose = False):
		'''
		Sets the original seeds of the diffusion. The genes must be nodes in
		the PPI.
		'''
		
		self.seeds['genes'] = self.ppi.getGenesInGraph(geneList)
		
	def setDiffusionTargets(self, geneList, verbose = False):
		'''
		Sets the desired targets of the diffusion. These are the genes that should
		be prioritised. Consider them only for evaluation!

		*IMPORTANT: run after setDiffusionGenes()
		'''

		self.targets = list(set(self.ppi.getGenesInGraph(geneList)) - set(self.seeds['genes']))


	def setDiffusionMims(self, mimList, verbose = False):
		'''
		Sets the original seeds of the diffusion. The mims are elements
		of the mim2gene object.
		'''

		#set the diffused diseases as seeds
		self.seeds['diseases'] = mimList

		genes = set()
		#add the possible genes 
		for mim in mimList:
			#retrieve genes per disease
			try:
				genes |= self.mm.getGenes(mim)
			except:
				pass

		#convert to list, required by setDiffusionGenes
		genes = list(genes)

		#retrieve the genes that are mappable in the PPI
		self.setDiffusionGenes(genes)

		if self.__verbose__ or verbose:
			print 'Obtained',len(self.seeds['genes']),'genes (seeds) from',len(mimList),'diseases'

	def diffuse(self, verbose = False):
		'''
		Diffuse information in the network
		'''

		#retrieves the similarity values from the diseases to be diffused
		geneWeights = self.buildGeneWeights()
		self.lastGeneWeights = geneWeights

		#gets the initial guess vector
		tmp = self.buildLabelMatrix(geneWeights = geneWeights)
		if isinstance(tmp, bool):
			return False
		

		kwargs = self.__staticDiffusionParameters__.copy()

		self.diffusionMethod.setParameters(**kwargs)
		
		self.lastDiffusionVector = self.diffusionMethod.diffuse(initialGuess = self.Y)
		self.predictionResult = PredictionResult(self.lastDiffusionVector, self.ppi.vertices, self.seeds['genes'], self.targets)

		return self.predictionResult

	def setDiffusionParameters(self, diffusionMethod, **kwargs):
		'''
		Set static parameters for the diffusion methods
		'''
		#set the currently used diffusion method label
		self.__diffusionMethod__ = diffusionMethod
		#set the static parameters to run the diffusion method
		self.__staticDiffusionParameters__ = kwargs

		#bind available diffusion methods
		if diffusionMethod == 'consistency':
			self.diffusionMethod = ConsistencyMethod(self.W, verbose=self.__verbose__)

	def scaleSimilarity(self, similarity):

		try:
			if self.similarityScale == 'sigmoid':
				return self.scaleSigmoid(similarity)
		except AttributeError:
			pass
		
		#default scaling is lineal
		return self.scaleLineal(similarity)	

	def scaleSigmoid(self, similarity, verbose = False):
		'''
		Scales similarity value to a sigmoid, which reaches 10 percent of its
		height at tenPercentSimilarity and 90 percent of its height at
		ninetyPercentSimilarity. 

		Call setSigmoidScale to change those values.
		'''

		if not hasattr(self, '__sigmoid__'):
			self.setSigmoidScale()

		#0.1 height is approx -2.2,  sig(-2.2) ~ 0.1
		#0.9 height is approx 2.2,  sig(2.2) ~ 0.9
		sigmoidScale = 4.4/(self.__sigmoid__['ninety']-self.__sigmoid__['ten'])
		#centre is obtained from the given values
		sigmoidHalf = (self.__sigmoid__['ninety']+self.__sigmoid__['ten'])/2

		#convert to sigmoid with halfpoint in similarity value
		scaledSimilarity = expit(sigmoidScale*similarity - sigmoidScale*sigmoidHalf)

		#shrink sigmoid to (0,height)
		scaledSimilarity *= self.__sigmoid__['height']

		return scaledSimilarity

	def setSigmoidScale(self, tenPercentSimilarity = 3.0, ninetyPercentSimilarity = 4.0, height = 0.5):
		'''
		Sets the parameters to make the sigmoid scaling.

		*needed for scaleSigmoid()
		'''
		self.__sigmoid__ = dict()

		self.__sigmoid__['ten'] = tenPercentSimilarity
		self.__sigmoid__['ninety'] = ninetyPercentSimilarity
		self.__sigmoid__['height'] = height

		self.similarityScale = 'sigmoid'

	def scaleLineal(self, similarity):
		'''
		Scales the similarity value in a lineal way in order to have 50
		percent of the height of the curve at fiftyPercentSimilarity. The
		height is a cutoff value, the resulting value is equal to the height
		after the curve intersects the height.
		'''

		if not hasattr(self, 'lineal'):
			self.setLinealScale()

		scaledSimilarity = similarity/(self.__lineal__['fifty']*2)
		if scaledSimilarity < self.__lineal__['height']:
			scaledSimilarity = self.__lineal__['height']
		if scaledSimilarity < 0:
			scaledSimilarity = 0

		return scaledSimilarity

	def setLinealScale(self, fiftyPercentSimilarity = 2.5, height = 0.5):
		'''
		Sets the parameters to make the lineal scaling.

		*needed for scaleLineal()
		'''

		self.__lineal__ = dict()

		self.__lineal__['fifty'] = fiftyPercentSimilarity
		self.__lineal__['height'] = height

		self.similarityScale = 'lineal'

	def buildGeneWeights(self, scale=True, verbose=False):
		'''
		Builds a dict with protein->weight according to the similarity between diseases and 
		the soruce mims.
		The similarities are found for diseases in the morbidmap that have an entry in the dissimDict.

		- sourceMims are the fixed mims that we're trying to compare
		
		'''
		geneWeights = defaultdict(int)		
		diseaseWeights = defaultdict(int)
		diseaseSims = defaultdict(int)


		#set of diseases that need weights (are in the morbid map, have
		#similarities and are not a diffusion seed)
		weightedDiseases = set(self.mm.getDiseases()) & self.ds.getDiseases()
		weightedDiseases -= set(self.seeds['diseases'])

		#get seeds with similarity
		seedsWithSimilarities = set(self.seeds['diseases']) & self.ds.getDiseases()

		#use the diseases that have similarity value from the morbidmap that are not diffused
		for mimB in weightedDiseases:
			#get the biggest similarity value for each disease against the diseases in the set
			for mimA in seedsWithSimilarities:

				#get the similarity value
				currentSim = self.ds.similarity(mimA, mimB)

				if not currentSim:
					#do not store similarities with value zero
					continue

				if scale:
					#scale the weight
					currentWeight = self.scaleSimilarity(currentSim)
				else:
					#do not scale the weight
					currentWeight = currentSim

				#if it is a new weight or a better weight
				if mimB not in diseaseWeights or diseaseWeights[mimB] < currentWeight:
					#scale the weights
					diseaseWeights[mimB] = currentWeight
					diseaseSims[mimB] = currentSim

			if mimB not in diseaseWeights:
				#if the mim had no similarities, do not add its proteins
				continue

			#set the weight for each disease protein equal to the highest similarity
			for gene in self.mm.getGenes(mimB):
				if gene not in geneWeights or geneWeights[gene] < diseaseWeights[mimB]:
					geneWeights[gene] = diseaseWeights[mimB]
	

		if self.__verbose__ or verbose:
			print 'Obtained similarities for',len(geneWeights),'proteins from',\
			len(diseaseWeights),'diseases' 

		return geneWeights

	def buildLabelMatrix(self, geneWeights=None, verbose=False, debug=False):
		'''
		Creates the matrix Y of initial labels from a set of proteins
		(labels should be between 0 and 1)

		- geneWeights is a dictionary: prot -> weight
			It is used to set the initial label for each prot to weight

		*if a protein is both in geneWeights and seeds['genes'], it will be labeled as a 1*
		'''

		#check if there is anything to build
		if len(geneWeights) == 0 and len(self.seeds['genes']) == 0:
			if self.__verbose__ or verbose:
				print 'Nothing to diffuse'
			return False


		#set the length of the initial prediction vector equal to the number
		#of genes in the PPI
		Y=np.zeros((len(self.ppi.vertices),1))

		similarGeneIndices=[]
		diffusedIndices = []

		#if there is information about similar diseases, genes from similar
		#diseases would get weights
		for gene in geneWeights:

			#get the position of this particular gene
			idx=self.ppi.getGeneIndex(gene)

			if not idx:
				continue

			#set partial weight
			Y[idx]=geneWeights[gene]
			similarGeneIndices.append(idx)
		if self.__verbose__ or verbose:
			print len(similarGeneIndices),'similar proteins from ',len(geneWeights),\
			'were mapped onto',self.ppi.name
		if debug:
			print 'Indices of the similar proteins',similarGeneIndices


		#if there are some seeds, set them as true labels for the diffusion
		for gene in self.seeds['genes']:
			#get the position of this particular gene
			idx=self.ppi.getGeneIndex(gene)

			if not idx:
				continue

			#set it as a label in the original guess vector
			Y[idx]=1
			#store the index
			diffusedIndices.append(idx)
		if self.__verbose__ or verbose:
			print len(diffusedIndices),'known proteins from ',len(self.seeds['genes']),\
			'were mapped onto',self.ppi.name
		if debug:
			print 'Indices of the true label proteins',diffusedIndices

		self.Y = Y

		return Y

	def getSeedOrdering(self, predictionResult):
		'''
		Retrieves the information about the seeds in the results
		'''

		return predictionResult.getGeneList(self.seeds['genes'])

	def setMeasure(self, measure):
		'''
		Sets the default measure for evaluation of the diffusion process
		'''
		self.measure = measure

	def evaluatePredictions(self, predictionResult=None, measure = None, measureParam=200):
		'''
		Evaluates the last predictions made

		measure is either 'mean' or 'roc'. The default is the mean position
		of the target gene within the predictions.
		'''

		if not predictionResult:
			predictionResult = self.predictionResult

		if not measure:
			measure = self.measure

		if len(self.targets) == 0:
			print 'Cannot evaluate'
			return float('inf')
		
		#this measure is done per individual target
		if measure == 'recall':
			m=Measures.L1oRecall(verbose=self.__verbose__)

			#if there is only one target, return the result as is
			if len(predictionResult.getTargetPositions()) == 1:
				return m.evaluate(predictionResult.getTargetPositions()[0],measureParam)

			#if there are many targets, return a list with all the individual evaluations
			else:
				out = []
				for pos in predictionResult.getTargetPositions():
					out.append(m.evaluate(pos,measureParam))
				return out

		m=Measures.VectorMean(verbose=self.__verbose__)
		if measure == 'roc':
			m=Measures.VectorROC(measureParam,verbose=self.__verbose__)

		return m.evaluate(predictionResult.getTargetPositions(),predictionResult.getTotalElements())

class DiffusionMethod(object):
	__metaclass__ = abc.ABCMeta

	@abc.abstractmethod
	def diffuse(self, initialGuess):
		'''
		Considering the diffusion can be modelled as Ax = b, where A is the 
		diffusion matrix, and b is the initial guess vector.

		This method must return the resulting vector x (as numpy array).

		initialGuess is the vector b

		*The diffuse method should run without parameters, they should be
		set in the setParameters() method.
		'''

	@abc.abstractmethod
	def setParameters(self, **kwargs):
		'''
		Establishes all the necessary parameters for the diffusion method.
		'''
	
	@abc.abstractmethod
	def getStaticParameters(self):
		'''
		Retrieves a list of dictionaries containing the static parameters' 
		name, range, and rangeType.
		[{'name': name, 'range': range,'rangeType': rangeType,'defaultValue':defaultValue}, ... ]

		'name' is the name of the variable to change
		'range' is a tuple when continuous, listing the first and last elements
			otherwise it is a list containing all the possible values.
		'rangeType' should be one of ['continuous', 'discrete']
		'defaultValue' is the value to use when nothing is provided

		'''

	@abc.abstractmethod
	def getName(self):
		'''
		Returns the name of the diffusion method
		'''

	@abc.abstractmethod
	def showParameters(self):
		'''
		Returns the list of parameters that must be passed.
		'''

class ConsistencyMethod(DiffusionMethod):
	'''
	Implementation of the Consistency Method (Zhou, NIPS 2004)
	'''
	def __init__(self, affinityMatrix, verbose = True):

		self.name = 'consistency'
		#static parameters
		self.alpha = int()

		#create affinity matrix
		self.S = self.__calculateDiffusionMatrix(affinityMatrix)
		#self.S = affinityMatrix
		#calculated parameters
		self.Y = []

		#constant parameters
		self.__verbose__ = verbose

	def diffuse(self, initialGuess=[]):
		'''
		Builds and solves the needed matrices for the linear system

		- The diffusion matrix S should be of shape D^(-1/2)*W*D^(-1/2)

		- The initial guess matrix Y is a binary matrix of size n*c, where c are the
		number of clusters. The positive examples of each cluster should have a 1.
		This variable can also be input as an integer, assuming that the first c
		entries of the diffusion matrix are examples of each of the c clusters.

		- alpha is a real number between 0 and 1, assigning a weight to the initial guess.
		'''

		if len(initialGuess) > 0:
			self.Y = initialGuess

		if self.S == []:
			raise UnboundLocalError('Missing parameters for diffusion')

		if self.__verbose__:
			print datetime.now(),'Building matrices'
			sys.stdout.flush()

		#if the initial guess is just a number, transform it to a vector of
		#that number of guesses
		if isinstance( self.Y, ( int, long ) ):
			#n will be the number of predictions made
			n=self.S.shape[0]
			self.Y = np.eye(n,self.Y)
		
		if self.__verbose__:
			print datetime.now(),'Solving linear system'
			sys.stdout.flush()
		
		#find the solution of the system
		F=la.cg(self.IalphaS,self.Y)
		#self.IalphaS is set within the setParameters method

		if self.__verbose__:
			print datetime.now(),'Done'
			sys.stdout.flush()
		return F[0]

	def __calculateDiffusionMatrix(self, weightMatrix):
		'''
		Creates a Diffusion matrix out of the affinity matrix W.

		The affinity matrix may be overriden if given as a parameter.
		'''

		self.W = weightMatrix

		#build D
		sums=self.W.sum(1) #sum every column per row

		#in case there are unconnected parts in the matrix
		indNonZeros = np.where(sums != 0)
		diagonalValues = np.zeros(sums.shape)

		diagonalValues[indNonZeros] = 1.0/np.sqrt(sums[indNonZeros])
		D = sparse.spdiags(diagonalValues.T,0,diagonalValues.shape[0],diagonalValues.shape[0])

		#build S
		S=D*self.W*D

		return S

	def setParameters(self, **kwargs):
		'''
		Set parameters for diffusion
		'''
		try:
			self.alpha = kwargs['alpha']
			#matrix to invert
			self.IalphaS=sparse.eye(self.S.shape[0])-self.alpha*self.S
		except KeyError:
			pass

		try:
			self.S = kwargs['diffusionMatrix']
		except KeyError:
			pass

		try:
			self.Y = kwargs['initialGuess']
		except KeyError:
			pass

	def getStaticParameters(self):
		'''
		Retrieves the static parameters name, range, and range type
		'''
		return [{'name':'alpha','range':(0.0,1.0),'rangeType':'continuous','defaultValue':0.75}]

	def getName(self):
		'''
		Returns the name of the diffusion method
		'''
		return self.name

	def showParameters(self):
		'''
		Returns list of the mandatory parameters
		'''
		return ['alpha','diffusionMatrix','initialGuess']


class PredictionResult:
	'''
	Helper object to show the results of a diffusion
	'''
	def __init__(self, weights, genes, seeds=[], targets=[]):

		positions = self.sort(weights)

		self.mapping(positions, genes)

		self.__seeds__ = seeds
		self.__targets__ = targets


	def mapping(self, positions, genes):
		'''
		Creates a dictionary with all the genes
		'''
		labelledDictionary = dict()
		orderedDictionary = dict()

		for position,pair in enumerate(positions):

			#retrieve the significant elements from the sorted list
			weight = pair[0]
			geneIdx = pair[1]
			#get the name of the gene
			label = genes[geneIdx]

			#fill the current gene's information
			values = dict()
			values['weight'] = weight
			values['position'] = position

			#set the label dictionary
			labelledDictionary[label] = values
			#set the order dictionary
			orderedDictionary[position] = label

		self.__labelledDict__ = labelledDictionary
		self.__orderedDict__ = orderedDictionary


	def getGeneInPosition(self, n):
		'''
		Retrieve the n-th ordered gene
		'''
		return self.__orderedDict__[n]

	def getFirstNGenes(self,n,ignoreSeeds=False):
		'''
		Retrieve the first n genes in order
		'''

		out = []

		if ignoreSeeds:
			seedPositions = self.getPositions(self.__seeds__, [], addName=False)
			n += len(seedPositions)

		for i in range(n):
			if ignoreSeeds:
				if i in seedPositions:
					#do not add a seed if we are ignoring seeds
					continue
			
			out.append(self.__orderedDict__[i])

		return out

	def getGenesInRange(self, lowerBound = 0, upperBound = False):
		'''
		Retrieve the genes inside a range.
		'''
		if not upperBound:
			upperBound = len(self.__orderedDict__)

		out = []
		for i in range(lowerBound,upperBound):
			out.append(self.getGeneInPosition(i))

		return out

	def getGene(self, gene):
		'''
		Retrieve the information about a single gene
		'''

		return self.__labelledDict__[gene]

	def getGeneList(self, geneList):
		'''
		Retrieves information about a list of genes
		'''
		out = []
		for gene in geneList:
			out.append(self.getGene(gene))

		return out

	def sort(self, weights):
		'''
		Sorts the weights and makes structures to access them
		'''
		try:
			positions = sorted([[val,idx] for idx,val in enumerate(weights)], reverse=True)
		except:
			positions = sorted([[val[0],idx] for idx,val in enumerate(weights)], reverse=True)

		return positions

	def getPositions(self, geneList, ignoreList = [], addName=False):
		'''
		Returns the positions of each gene
		'''

		# if needed, subtract the number of ignored genes listed before the query genes
		ignores = [(self.__labelledDict__[x]['position']) for x in ignoreList]

		if addName:
			return [(self.__labelledDict__[x]['position']
					-len([y for y in ignores if self.__labelledDict__[x]['position'] > y])
					,x) for x in geneList]
		else:
			return [self.__labelledDict__[x]['position'] 
					-len([y for y in ignores if self.__labelledDict__[x]['position'] > y])
					for x in geneList]

	def getTargetPositions(self,ignoreSeeds = True, addName = False):
		'''
		Returns the position of the seed genes

		ignore seeds will pretend the seeds are non existent in the result list.
		'''

		if ignoreSeeds:
			return self.getPositions(self.__targets__, self.__seeds__,addName=addName)
		else:
			return self.getPositions(self.__targets__, [], addName=addName)

	def getBelongingList(self, ignoreSeeds = True):
		'''
		Returns a boolean vector. The positions of the genes correspond to the sorted
		mapping. The value will be True if the gene is a target.
		'''

		#true if the label is a target
		belonging = [self.__orderedDict__[i] in self.__targets__ for i in range(len(self.__orderedDict__)) ]

		if ignoreSeeds:
			#ignore the positions of the seeds
			ignores = [(self.__labelledDict__[x]['position']) for x in self.__seeds__]
			#temp copy of results
			tmp = [x for e,x in enumerate(belonging) if e not in ignores]
			#establish proper return variable
			belonging = tmp

		return belonging

	def getTotalElements(self):
		'''
		Returns the total number of elements in the listing
		'''
		return len(self.__orderedDict__)


